@@ -16,7 +16,8 @@ use crate::{
     Error,
     Result,
     crypto::{
-        asymmetric::{KeyPair, Decryptor, Signer},
+        asymmetric::KeyPair,
+        backend::interface::Asymmetric,
         mem::Protected,
         mpi::{self, MPI, ProtectedMPI, PublicKey},
         SessionKey,
@@ -32,6 +33,115 @@ use crate::{
     },
 };
 
+impl Asymmetric for super::Backend {
+    fn supports_algo(algo: PublicKeyAlgorithm) -> bool {
+        use PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        match algo {
+            RSAEncryptSign | RSAEncrypt | RSASign | DSA | ECDH | ECDSA | EdDSA |
+            ElGamalEncrypt | ElGamalEncryptSign
+                => true,
+            Private(_) | Unknown(_)
+                => false,
+        }
+    }
+
+    fn supports_curve(curve: &Curve) -> bool {
+        use Curve::*;
+        match curve {
+            NistP256 | NistP384 | NistP521 | Ed25519 | Cv25519 |
+            BrainpoolP256 | BrainpoolP512
+                => true,
+            Unknown(_) if curve.is_brainpoolp384() // XXX
+                => true,
+            Unknown(_)
+                => false,
+        }
+    }
+
+    fn x25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        let mut rng = RandomNumberGenerator::new_userspace()?;
+        let secret = Privkey::create("Curve25519", "", &mut rng)?;
+        let mut public = [0u8; 32];
+        public.copy_from_slice(&secret.pubkey()?.get_x25519_key()?);
+        let mut secret: Protected = secret.get_x25519_key()?.into();
+
+        // Clamp the scalar.  X25519 does the clamping implicitly, but
+        // OpenPGP's ECDH over Curve25519 requires the secret to be
+        // clamped.
+        secret[0] &= 0b1111_1000;
+        secret[31] &= !0b1000_0000;
+        secret[31] |= 0b0100_0000;
+
+        Ok((secret, public))
+    }
+
+    fn x25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        let secret = Privkey::load_x25519(secret)?;
+        Ok(<[u8; 32]>::try_from(&secret.pubkey()?.get_x25519_key()?[..])?)
+    }
+
+    fn x25519_shared_point(secret: &Protected, public: &[u8; 32])
+                           -> Result<Protected> {
+        let secret = Privkey::load_x25519(&secret)?;
+        Ok(secret.agree(public, 32, b"", "Raw")?.into())
+    }
+
+    fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        let mut rng = RandomNumberGenerator::new_userspace()?;
+        let secret = Privkey::create("Ed25519", "", &mut rng)?;
+        let (public, secret) = secret.get_ed25519_key()?;
+        Ok((secret.into(), public.as_slice().try_into()?))
+    }
+
+    fn ed25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        let secret = Privkey::load_ed25519(secret)?;
+        let (public, secret) = secret.get_ed25519_key()?;
+        let _ = Protected::from(secret); // Securely dispose.
+        Ok(public.as_slice().try_into()?)
+    }
+
+    fn ed25519_sign(secret: &Protected, _public: &[u8; 32], digest: &[u8])
+                    -> Result<[u8; 64]> {
+        let mut rng = RandomNumberGenerator::new_userspace()?;
+        let secret = Privkey::load_ed25519(&secret)?;
+        Ok(secret.sign(digest, "", &mut rng)?.as_slice().try_into()?)
+    }
+
+    fn ed25519_verify(public: &[u8; 32], digest: &[u8], signature: &[u8; 64])
+                      -> Result<bool> {
+        let pk = Pubkey::load_ed25519(public)?;
+        Ok(pk.verify(digest, signature, "")?)
+    }
+
+    fn dsa_generate_key(p_bits: usize)
+                        -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
+    {
+        let mut rng = RandomNumberGenerator::new_userspace()?;
+        let q_bits = if p_bits <= 1024 { 160 } else { 256 };
+        let secret = Privkey::create_dsa(p_bits, q_bits, &mut rng)?;
+        let public = secret.pubkey()?;
+        Ok((public.get_field("p")?.try_into()?,
+            public.get_field("q")?.try_into()?,
+            public.get_field("g")?.try_into()?,
+            public.get_field("y")?.try_into()?,
+            secret.get_field("x")?.try_into()?))
+    }
+
+    fn elgamal_generate_key(p_bits: usize)
+                            -> Result<(MPI, MPI, MPI, ProtectedMPI)>
+    {
+        let mut rng = RandomNumberGenerator::new_userspace()?;
+        let q_bits = if p_bits <= 1024 { 160 } else { 256 };
+        let secret = Privkey::create_elgamal(p_bits, q_bits, &mut rng)?;
+        let public = secret.pubkey()?;
+        Ok((public.get_field("p")?.try_into()?,
+            public.get_field("g")?.try_into()?,
+            public.get_field("y")?.try_into()?,
+            secret.get_field("x")?.try_into()?))
+    }
+}
+
 // CONFIDENTIALITY: Botan clears the MPIs after use.
 impl TryFrom<&ProtectedMPI> for botan::MPI {
     type Error = anyhow::Error;
@@ -75,21 +185,19 @@ impl TryFrom<botan::MPI> for MPI {
     }
 }
 
-impl Signer for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8])
-            -> Result<mpi::Signature>
+impl KeyPair {
+    pub(crate) fn sign_backend(&self,
+                               secret: &mpi::SecretKeyMaterial,
+                               hash_algo: HashAlgorithm,
+                               digest: &[u8])
+                               -> Result<mpi::Signature>
     {
         use crate::PublicKeyAlgorithm::*;
 
         let mut rng = RandomNumberGenerator::new_userspace()?;
 
-        self.secret().map(|secret| {
-            #[allow(deprecated)]
-            match (self.public().pk_algo(), self.public().mpis(), secret)
+        #[allow(deprecated)]
+        match (self.public().pk_algo(), self.public().mpis(), secret)
         {
             (RSASign,
              PublicKey::RSA { e, .. },
@@ -129,30 +237,6 @@ impl Signer for KeyPair {
                 })
             },
 
-            (EdDSA,
-             PublicKey::EdDSA { curve, .. },
-             mpi::SecretKeyMaterial::EdDSA { scalar }) => match curve {
-                Curve::Ed25519 => {
-                    let size = 32;
-                    let scalar = scalar.value_padded(size);
-                    let secret = Privkey::load_ed25519(&scalar)?;
-                    let sig = secret.sign(digest, "", &mut rng)?;
-
-                    if sig.len() != size * 2 {
-                        return Err(Error::MalformedMPI(
-                            format!("Expected signature with length {}, got {}",
-                                    size * 2, sig.len())).into());
-                    }
-
-                    Ok(mpi::Signature::EdDSA {
-                        r: MPI::new(&sig[..size]),
-                        s: MPI::new(&sig[size..]),
-                    })
-                },
-                _ => Err(
-                    Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
-
             (ECDSA,
              PublicKey::ECDSA { curve, .. },
              mpi::SecretKeyMaterial::ECDSA { scalar }) => {
@@ -177,17 +261,13 @@ impl Signer for KeyPair {
                 "unsupported combination of algorithm {:?}, key {:?}, \
                  and secret key {:?}",
                 pk_algo, self.public(), self.secret())).into()),
-        }})
+        }
     }
 }
 
-impl Decryptor for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn decrypt(&mut self, ciphertext: &mpi::Ciphertext,
-               _plaintext_len: Option<usize>)
+impl KeyPair {
+    pub(crate) fn decrypt_backend(&self, secret: &mpi::SecretKeyMaterial, ciphertext: &mpi::Ciphertext,
+                                  plaintext_len: Option<usize>)
                -> Result<SessionKey>
     {
         fn bad(e: impl ToString) -> anyhow::Error {
@@ -195,9 +275,7 @@ impl Decryptor for KeyPair {
             Error::MalformedMessage(e.to_string()).into()
         }
 
-        self.secret().map(
-            |secret| Ok(match (self.public().mpis(), secret, ciphertext)
-        {
+        Ok(match (self.public().mpis(), secret, ciphertext) {
             (PublicKey::RSA { e, .. },
              mpi::SecretKeyMaterial::RSA { p, q, .. },
              mpi::Ciphertext::RSA { c }) => {
@@ -229,21 +307,22 @@ impl Decryptor for KeyPair {
             (PublicKey::ECDH{ .. },
              mpi::SecretKeyMaterial::ECDH { .. },
              mpi::Ciphertext::ECDH { .. }) =>
-                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext)?,
+                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext,
+                                             plaintext_len)?,
 
             (public, secret, ciphertext) =>
                 return Err(Error::InvalidOperation(format!(
                     "unsupported combination of key pair {:?}/{:?} \
                      and ciphertext {:?}",
                     public, secret, ciphertext)).into()),
-        }))
+        })
     }
 }
 
 
 impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     /// Encrypts the given data with this key.
-    pub fn encrypt(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
+    pub(crate) fn encrypt_backend(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
         use crate::PublicKeyAlgorithm::*;
 
         #[allow(deprecated)]
@@ -313,7 +392,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     }
 
     /// Verifies the given signature.
-    pub fn verify(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
+    pub(crate) fn verify_backend(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
                   digest: &[u8]) -> Result<()>
     {
         use crate::crypto::mpi::Signature;
@@ -345,30 +424,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                 let truncated_digest = &digest[..size.min(digest.len())];
                 pk.verify(truncated_digest, &sig, "Raw").unwrap()
             },
-            (PublicKey::EdDSA { curve, q }, Signature::EdDSA { r, s }) =>
-              match curve {
-                Curve::Ed25519 => {
-                    if q.value().get(0).map(|&b| b != 0x40).unwrap_or(true) {
-                        return Err(Error::MalformedPacket(
-                            "Invalid point encoding".into()).into());
-                    }
-
-                    // OpenPGP encodes R and S separately, but our
-                    // cryptographic library expects them to be
-                    // concatenated.
-                    let mut sig = Vec::with_capacity(64);
-
-                    // We need to zero-pad them at the front, because
-                    // the MPI encoding drops leading zero bytes.
-                    sig.extend_from_slice(&r.value_padded(32).map_err(bad)?);
-                    sig.extend_from_slice(&s.value_padded(32).map_err(bad)?);
-
-                    let pk = Pubkey::load_ed25519(&q.value()[1..])?;
-                    pk.verify(digest, &sig, "")?
-                },
-                _ => return
-                    Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
             (PublicKey::ECDSA { curve, q }, Signature::ECDSA { s, r }) =>
             {
                 // OpenPGP encodes R and S separately, but our
@@ -404,37 +459,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
 impl<R> Key4<SecretParts, R>
     where R: key::KeyRole,
 {
-    pub(crate) fn derive_cv25519_public_key(private_key: &Protected) -> Result<[u8; 32]>
-    {
-        let secret = Privkey::load_x25519(private_key)?;
-
-        Ok(<[u8; 32]>::try_from(&secret.pubkey()?.get_x25519_key()?[..])?)
-    }
-
-    /// Creates a new OpenPGP secret key packet for an existing Ed25519 key.
-    ///
-    /// The ECDH key will use hash algorithm `hash` and symmetric
-    /// algorithm `sym`.  If one or both are `None` secure defaults
-    /// will be used.  The key will have it's creation date set to
-    /// `ctime` or the current time if `None` is given.
-    pub fn import_secret_ed25519<T>(private_key: &[u8], ctime: T)
-        -> Result<Self> where T: Into<Option<SystemTime>>
-    {
-        let secret = Privkey::load_ed25519(private_key)?;
-        let (public, secret) = secret.get_ed25519_key()?;
-
-        Self::with_secret(
-            ctime.into().unwrap_or_else(crate::now),
-            PublicKeyAlgorithm::EdDSA,
-            mpi::PublicKey::EdDSA {
-                curve: Curve::Ed25519,
-                q: MPI::new_compressed_point(&public),
-            },
-            mpi::SecretKeyMaterial::EdDSA {
-                scalar: secret.into(),
-            }.into())
-    }
-
     /// Creates a new OpenPGP public key packet for an existing RSA key.
     ///
     /// The RSA key will use public exponent `e` and modulo `n`. The key will
@@ -462,7 +486,7 @@ impl<R> Key4<SecretParts, R>
             public, secret.into())
     }
 
-    /// Generates a new RSA key with a public modulos of size `bits`.
+    /// Generates a new RSA key with a public modulus of size `bits`.
     pub fn generate_rsa(bits: usize) -> Result<Self> {
         let mut rng = RandomNumberGenerator::new_userspace()?;
         let secret = Privkey::create("RSA", &format!("{}", bits), &mut rng)?;
