@@ -6,9 +6,10 @@ use std::convert::TryInto;
 
 use crate::{Error, Result};
 
-use crate::crypto::asymmetric::{Decryptor, KeyPair, Signer};
+use crate::crypto::asymmetric::KeyPair;
+use crate::crypto::backend::interface::Asymmetric;
 use crate::crypto::mem::Protected;
-use crate::crypto::mpi;
+use crate::crypto::mpi::{self, MPI, ProtectedMPI};
 use crate::crypto::SessionKey;
 use crate::crypto::{pad, pad_at_least, pad_truncating};
 use crate::packet::key::{Key4, SecretParts};
@@ -21,12 +22,154 @@ use win_crypto_ng as cng;
 
 const CURVE25519_SIZE: usize = 32;
 
-impl Signer for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
+impl TryFrom<&Protected> for Box<ed25519_dalek::SigningKey> {
+    type Error = anyhow::Error;
+
+    fn try_from(value: &Protected) -> Result<Self> {
+        if value.len() != ed25519_dalek::SECRET_KEY_LENGTH {
+            return Err(crate::Error::InvalidArgument(
+                "Bad Ed25519 secret length".into()).into());
+        }
+        Ok(Box::new(ed25519_dalek::SigningKey::from_bytes(
+            value.as_ref().try_into().map_err(
+                |e: std::array::TryFromSliceError| {
+                    Error::InvalidKey(e.to_string())
+                })?)))
+    }
+}
+
+impl Asymmetric for super::Backend {
+    fn supports_algo(algo: PublicKeyAlgorithm) -> bool {
+        use PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        match algo {
+            RSAEncryptSign | RSAEncrypt | RSASign | DSA | ECDH | ECDSA | EdDSA
+                => true,
+            ElGamalEncrypt | ElGamalEncryptSign | Private(_) | Unknown(_)
+                => false,
+        }
     }
 
-    fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8]) -> Result<mpi::Signature> {
+    fn supports_curve(curve: &Curve) -> bool {
+        use Curve::*;
+        match curve {
+            NistP256 | NistP384 | NistP521 | Ed25519 | Cv25519
+                => true,
+            BrainpoolP256 | BrainpoolP512 | Unknown(_)
+                => false,
+        }
+    }
+
+    fn x25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        use cng::asymmetric::{Ecdh, AsymmetricKey, Export};
+        use cng::asymmetric::ecc::Curve25519;
+
+        let pair =
+            AsymmetricKey::builder(Ecdh(Curve25519)).build()?.export()?;
+
+        let mut public = [0u8; 32];
+        public.copy_from_slice(pair.x());
+
+        Ok((pair.d().into(), public))
+    }
+
+    fn x25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        use cng::asymmetric::{AsymmetricAlgorithm, AsymmetricAlgorithmId, Ecdh,
+                              Private, AsymmetricKey, Export};
+        use cng::asymmetric::ecc::{Curve25519, NamedCurve};
+
+        let provider = AsymmetricAlgorithm::open(
+            AsymmetricAlgorithmId::Ecdh(NamedCurve::Curve25519)
+        )?;
+        let key = AsymmetricKey::<Ecdh<Curve25519>, Private>::import_from_parts(
+            &provider,
+            secret,
+        )?;
+        Ok(<[u8; 32]>::try_from(&key.export()?.x()[..])?)
+    }
+
+    fn x25519_shared_point(secret: &Protected, public: &[u8; 32])
+                           -> Result<Protected> {
+        use cng::asymmetric::{Ecdh, AsymmetricKey, Public, Private,
+                              AsymmetricAlgorithm, AsymmetricAlgorithmId};
+        use cng::asymmetric::agreement::secret_agreement;
+        use cng::asymmetric::ecc::{NamedCurve, Curve25519};
+
+        let provider =
+            AsymmetricAlgorithm::open(
+                AsymmetricAlgorithmId::Ecdh(NamedCurve::Curve25519))?;
+        let public =
+            AsymmetricKey::<Ecdh<Curve25519>, Public>::import_from_parts(
+                &provider,
+                public,
+            )?;
+        let secret =
+            AsymmetricKey::<Ecdh<Curve25519>, Private>::import_from_parts(
+                &provider,
+                secret,
+            )?;
+
+        let shared = secret_agreement(&secret, &public)?;
+        let mut shared = Protected::from(shared.derive_raw()?);
+        // Returned secret is little-endian, flip it to big-endian
+        shared.reverse();
+        Ok(shared)
+    }
+
+    fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        let mut rng = cng::random::RandomNumberGenerator::system_preferred();
+        let pair = ed25519_dalek::SigningKey::generate(&mut rng);
+        Ok((pair.to_bytes().into(), pair.verifying_key().to_bytes()))
+    }
+
+    fn ed25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        use ed25519_dalek::SigningKey;
+        let secret: Box<SigningKey> = secret.try_into()?;
+        let public = secret.verifying_key();
+        Ok(public.to_bytes())
+    }
+
+    fn ed25519_sign(secret: &Protected, _public: &[u8; 32], digest: &[u8])
+                    -> Result<[u8; 64]> {
+        use ed25519_dalek::{SigningKey, Signer};
+        let secret: Box<SigningKey> = secret.try_into()?;
+        Ok(secret.sign(digest).to_bytes().try_into()?)
+    }
+
+    fn ed25519_verify(public: &[u8; 32], digest: &[u8], signature: &[u8; 64])
+                      -> Result<bool> {
+        use ed25519_dalek::{VerifyingKey, Verifier, Signature};
+
+        let public = VerifyingKey::from_bytes(public).map_err(|e| {
+            Error::InvalidKey(e.to_string())
+        })?;
+        let signature = signature.as_ref().try_into().map_err(|e: std::array::TryFromSliceError| {
+            Error::InvalidArgument(e.to_string())
+        })?;
+
+        let signature = Signature::from_bytes(signature);
+        Ok(public.verify(digest, &signature).is_ok())
+    }
+
+    fn dsa_generate_key(p_bits: usize)
+                        -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
+    {
+        // XXX: DSA key generation needs fixes upstream, or at least I
+        // didn't figure out how to do that properly, see
+        // https://github.com/emgre/win-crypto-ng/issues/47
+        let _ = p_bits;
+        Err(Error::UnsupportedPublicKeyAlgorithm(
+            PublicKeyAlgorithm::ElGamalEncrypt).into())
+    }
+}
+
+impl KeyPair {
+    pub(crate) fn sign_backend(&self,
+                               secret: &mpi::SecretKeyMaterial,
+                               hash_algo: HashAlgorithm,
+                               digest: &[u8])
+                               -> Result<mpi::Signature>
+    {
         use cng::asymmetric::{AsymmetricAlgorithm, AsymmetricAlgorithmId};
         use cng::asymmetric::{AsymmetricKey, Private, Rsa};
         use cng::asymmetric::signature::{Signer, SignaturePadding};
@@ -35,8 +178,7 @@ impl Signer for KeyPair {
         use cng::asymmetric::ecc::NamedCurve;
 
         #[allow(deprecated)]
-        self.secret().map(|secret| {
-            Ok(match (self.public().pk_algo(), self.public().mpis(), secret) {
+        Ok(match (self.public().pk_algo(), self.public().mpis(), secret) {
                 (PublicKeyAlgorithm::RSAEncryptSign,
                     &mpi::PublicKey::RSA { ref e, ref n },
                     &mpi::SecretKeyMaterial::RSA { ref p, ref q, ref d, .. }) |
@@ -128,47 +270,7 @@ impl Signer for KeyPair {
                         s: mpi::MPI::new(s),
                     }
                 },
-                (
-                    PublicKeyAlgorithm::EdDSA,
-                    mpi::PublicKey::EdDSA { curve, q },
-                    mpi::SecretKeyMaterial::EdDSA { scalar },
-                ) => match curve {
-                    Curve::Ed25519 => {
-                        // CNG doesn't support EdDSA, use ed25519-dalek instead
-                        use ed25519_dalek::{Keypair, Signer};
-                        use ed25519_dalek::{PUBLIC_KEY_LENGTH, SECRET_KEY_LENGTH};
 
-                        let (public, ..) = q.decode_point(&Curve::Ed25519)?;
-                        // MPI::decode_point ensures we got the right length.
-                        assert_eq!(public.len(), PUBLIC_KEY_LENGTH);
-
-                        // It's expected for the private key to be exactly
-                        // SECRET_KEY_LENGTH bytes long but OpenPGP allows leading
-                        // zeros to be stripped.
-                        // Padding has to be unconditional; otherwise we have a
-                        // secret-dependent branch.
-                        let mut keypair = Protected::from(
-                            vec![0u8; SECRET_KEY_LENGTH + PUBLIC_KEY_LENGTH]
-                        );
-                        keypair[..SECRET_KEY_LENGTH]
-                            .copy_from_slice(
-                                &scalar.value_padded(SECRET_KEY_LENGTH));
-                        keypair[SECRET_KEY_LENGTH..]
-                            .copy_from_slice(&public);
-                        let pair = Keypair::from_bytes(&keypair).unwrap();
-
-                        let sig = pair.sign(digest).to_bytes();
-
-                        // https://tools.ietf.org/html/rfc8032#section-5.1.6
-                        let (r, s) = sig.split_at(sig.len() / 2);
-                        mpi::Signature::EdDSA {
-                            r: mpi::MPI::new(r),
-                            s: mpi::MPI::new(s),
-                        }
-                    },
-                    _ => return Err(
-                        Error::UnsupportedEllipticCurve(curve.clone()).into()),
-                },
                 (PublicKeyAlgorithm::DSA,
                     mpi:: PublicKey::DSA { y, p, q, g },
                     mpi::SecretKeyMaterial::DSA { x },
@@ -293,27 +395,20 @@ impl Signer for KeyPair {
                     "unsupported combination of algorithm {:?}, key {:?}, \
                      and secret key {:?}",
                     pk_algo, self.public(), self.secret())))?,
-            })
         })
     }
 }
 
-impl Decryptor for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    /// Creates a signature over the `digest` produced by `hash_algo`.
-    fn decrypt(
-        &mut self,
+impl KeyPair {
+    pub(crate) fn decrypt_backend(
+        &self,
+        secret: &mpi::SecretKeyMaterial,
         ciphertext: &mpi::Ciphertext,
         plaintext_len: Option<usize>,
     ) -> Result<SessionKey> {
         use crate::PublicKeyAlgorithm::*;
 
-        self.secret().map(
-            |secret| Ok(match (self.public().mpis(), secret, ciphertext)
-        {
+        Ok(match (self.public().mpis(), secret, ciphertext) {
             (mpi::PublicKey::RSA { ref e, ref n },
              mpi::SecretKeyMaterial::RSA { ref p, ref q, ref d, .. },
              mpi::Ciphertext::RSA { ref c }) => {
@@ -354,20 +449,21 @@ impl Decryptor for KeyPair {
             (mpi::PublicKey::ECDH{ .. },
              mpi::SecretKeyMaterial::ECDH { .. },
              mpi::Ciphertext::ECDH { .. }) =>
-                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext)?,
+                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext,
+                                             plaintext_len)?,
 
             (public, secret, ciphertext) =>
                 return Err(Error::InvalidOperation(format!(
                     "unsupported combination of key pair {:?}/{:?} \
                      and ciphertext {:?}",
                     public, secret, ciphertext)).into()),
-        }))
+        })
     }
 }
 
 impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     /// Encrypts the given data with this key.
-    pub fn encrypt(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
+    pub(crate) fn encrypt_backend(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
         use cng::asymmetric::{AsymmetricAlgorithm, AsymmetricAlgorithmId};
         use cng::asymmetric::{AsymmetricKey, Public, Rsa};
         use cng::key_blob::RsaKeyPublicPayload;
@@ -426,7 +522,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     }
 
     /// Verifies the given signature.
-    pub fn verify(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
+    pub(crate) fn verify_backend(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
                   digest: &[u8]) -> Result<()> {
         use cng::asymmetric::{AsymmetricAlgorithm, AsymmetricAlgorithmId};
         use cng::asymmetric::{AsymmetricKey, Public, Rsa};
@@ -640,42 +736,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                         Error::UnsupportedEllipticCurve(curve.clone()).into()),
                 }
             },
-            (mpi::PublicKey::EdDSA { curve, q }, mpi::Signature::EdDSA { r, s })
-                => match curve
-            {
-                Curve::Ed25519 => {
-                    // CNG doesn't support EdDSA, use ed25519-dalek instead
-                    use ed25519_dalek::{PublicKey, Signature, SIGNATURE_LENGTH};
-                    use ed25519_dalek::{Verifier};
-
-                    let (public, ..) = q.decode_point(&Curve::Ed25519)?;
-                    assert_eq!(public.len(), 32);
-
-                    let key = PublicKey::from_bytes(public).map_err(|e| {
-                        Error::InvalidKey(e.to_string())
-                    })?;
-
-                    // ed25519 expects full-sized signatures but OpenPGP allows
-                    // for stripped leading zeroes, pad each part with zeroes.
-                    let mut sig_bytes = [0u8; SIGNATURE_LENGTH];
-
-                    // We need to zero-pad them at the front, because
-                    // the MPI encoding drops leading zero bytes.
-                    let half = SIGNATURE_LENGTH / 2;
-                    sig_bytes[..half].copy_from_slice(
-                        &r.value_padded(half).map_err(bad)?);
-                    sig_bytes[half..].copy_from_slice(
-                        &s.value_padded(half).map_err(bad)?);
-
-                    let signature = Signature::from(sig_bytes);
-
-                    key.verify(digest, &signature)
-                        .map(|_| true)
-                        .map_err(|e| Error::BadSignature(e.to_string()))?
-                },
-                _ => return Err(
-                    Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
             _ => return Err(Error::MalformedPacket(format!(
                 "unsupported combination of key {} and signature {:?}.",
                 self.pk_algo(), sig)).into()),
@@ -693,56 +753,6 @@ impl<R> Key4<SecretParts, R>
 where
     R: key::KeyRole,
 {
-    pub(crate) fn derive_cv25519_public_key(private_key: &Protected) -> Result<[u8; 32]>
-    {
-        use cng::asymmetric::{AsymmetricAlgorithm, AsymmetricAlgorithmId, Ecdh, Private};
-        use cng::asymmetric::{AsymmetricKey, Export};
-        use cng::asymmetric::ecc::{Curve25519, NamedCurve};
-
-        let provider = AsymmetricAlgorithm::open(
-            AsymmetricAlgorithmId::Ecdh(NamedCurve::Curve25519)
-        )?;
-        let key = AsymmetricKey::<Ecdh<Curve25519>, Private>::import_from_parts(
-            &provider,
-            private_key
-        )?;
-        Ok(<[u8; 32]>::try_from(&key.export()?.x()[..])?)
-    }
-
-    /// Creates a new OpenPGP secret key packet for an existing Ed25519 key.
-    ///
-    /// The key will have it's creation date set to `ctime` or the current time
-    /// if `None` is given.
-    pub fn import_secret_ed25519<T>(private_key: &[u8], ctime: T) -> Result<Self>
-    where
-        T: Into<Option<SystemTime>>,
-    {
-        // CNG doesn't support EdDSA, use ed25519-dalek instead
-        use ed25519_dalek::{PublicKey, SecretKey};
-
-        let private = SecretKey::from_bytes(private_key).map_err(|e| {
-            Error::InvalidKey(e.to_string())
-        })?;
-
-        // Mark MPI as compressed point with 0x40 prefix. See
-        // https://tools.ietf.org/html/draft-ietf-openpgp-rfc4880bis-07#section-13.2.
-        let mut public = [0u8; 1 + CURVE25519_SIZE];
-        public[0] = 0x40;
-        public[1..].copy_from_slice(Into::<PublicKey>::into(&private).as_bytes());
-
-        Self::with_secret(
-            ctime.into().unwrap_or_else(crate::now),
-            PublicKeyAlgorithm::EdDSA,
-            mpi::PublicKey::EdDSA {
-                curve: Curve::Ed25519,
-                q: mpi::MPI::new(&public)
-            },
-            mpi::SecretKeyMaterial::EdDSA {
-                scalar: private_key.into(),
-            }.into()
-        )
-    }
-
     /// Creates a new OpenPGP public key packet for an existing RSA key.
     ///
     /// The RSA key will use public exponent `e` and modulo `n`. The key will
@@ -786,7 +796,7 @@ where
         )
     }
 
-    /// Generates a new RSA key with a public modulos of size `bits`.
+    /// Generates a new RSA key with a public modulus of size `bits`.
     pub fn generate_rsa(bits: usize) -> Result<Self> {
         use win_crypto_ng::asymmetric::{AsymmetricKey, Rsa};
 
@@ -911,18 +921,18 @@ where
             },
             (Curve::Ed25519, true) => {
                 // CNG doesn't support EdDSA, use ed25519-dalek instead
-                use ed25519_dalek::Keypair;
+                use ed25519_dalek::SigningKey;
 
                 let mut rng = cng::random::RandomNumberGenerator::system_preferred();
-                let Keypair { public, secret } = Keypair::generate(&mut rng);
+                let key = SigningKey::generate(&mut rng);
 
-                let secret: Protected = secret.as_bytes().as_ref().into();
+                let secret: Protected = key.to_bytes().as_ref().into();
 
                 // Mark MPI as compressed point with 0x40 prefix. See
                 // https://tools.ietf.org/html/draft-ietf-openpgp-rfc4880bis-07#section-13.2.
                 let mut compressed_public = [0u8; 1 + CURVE25519_SIZE];
                 compressed_public[0] = 0x40;
-                compressed_public[1..].copy_from_slice(public.as_bytes());
+                compressed_public[1..].copy_from_slice(key.verifying_key().as_bytes());
 
                 (
                     EdDSA,
