@@ -1,34 +1,126 @@
 //! Hold the implementation of [`Signer`] and [`Decryptor`] for [`KeyPair`].
 //!
-//! [`Signer`]: super::super::asymmetric::Signer
-//! [`Decryptor`]: super::super::asymmetric::Decryptor
-//! [`KeyPair`]: super::super::asymmetric::KeyPair
+//! [`Signer`]: crate::crypto::Signer
+//! [`Decryptor`]: crate::crypto::Decryptor
+//! [`KeyPair`]: crate::crypto::KeyPair
 
 use nettle::{curve25519, ecc, ecdh, ecdsa, ed25519, dsa, rsa, random::Yarrow};
 
 use crate::{Error, Result};
 
 use crate::packet::{key, Key};
-use crate::crypto::asymmetric::{KeyPair, Decryptor, Signer};
-use crate::crypto::mpi::{self, MPI, PublicKey};
+use crate::crypto::asymmetric::KeyPair;
+use crate::crypto::backend::interface::Asymmetric;
+use crate::crypto::mpi::{self, MPI, ProtectedMPI, PublicKey};
 use crate::crypto::SessionKey;
 use crate::types::{Curve, HashAlgorithm};
 
-impl Signer for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
+impl Asymmetric for super::Backend {
+    fn supports_algo(algo: PublicKeyAlgorithm) -> bool {
+        use PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        match algo {
+            RSAEncryptSign | RSAEncrypt | RSASign | DSA | ECDH | ECDSA | EdDSA
+                => true,
+            ElGamalEncrypt | ElGamalEncryptSign | Private(_) | Unknown(_)
+                => false,
+        }
     }
 
-    fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8])
-            -> Result<mpi::Signature>
+    fn supports_curve(curve: &Curve) -> bool {
+        use Curve::*;
+        match curve {
+            NistP256 | NistP384 | NistP521 | Ed25519 | Cv25519
+                => true,
+            BrainpoolP256 | BrainpoolP512 | Unknown(_)
+                => false,
+        }
+    }
+
+    fn x25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        debug_assert_eq!(curve25519::CURVE25519_SIZE, 32);
+        let mut rng = Yarrow::default();
+        let secret = curve25519::private_key(&mut rng);
+        let mut public = [0; 32];
+        curve25519::mul_g(&mut public, &secret)?;
+        Ok((secret.into(), public))
+    }
+
+    fn x25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        debug_assert_eq!(curve25519::CURVE25519_SIZE, 32);
+        let mut public = [0; 32];
+        curve25519::mul_g(&mut public, secret)?;
+        Ok(public)
+    }
+
+    fn x25519_shared_point(secret: &Protected, public: &[u8; 32])
+                           -> Result<Protected> {
+        debug_assert_eq!(curve25519::CURVE25519_SIZE, 32);
+        let mut s: Protected = vec![0; 32].into();
+        curve25519::mul(&mut s, secret, public)?;
+        Ok(s)
+    }
+
+    fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        debug_assert_eq!(ed25519::ED25519_KEY_SIZE, 32);
+        let mut rng = Yarrow::default();
+        let mut public = [0; 32];
+        let secret: Protected =
+            ed25519::private_key(&mut rng).into();
+        ed25519::public_key(&mut public, &secret)?;
+        Ok((secret, public))
+    }
+
+    fn ed25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        debug_assert_eq!(ed25519::ED25519_KEY_SIZE, 32);
+        let mut public = [0; 32];
+        ed25519::public_key(&mut public, secret)?;
+        Ok(public)
+    }
+
+    fn ed25519_sign(secret: &Protected, public: &[u8; 32], digest: &[u8])
+                    -> Result<[u8; 64]> {
+        debug_assert_eq!(ed25519::ED25519_KEY_SIZE, 32);
+        debug_assert_eq!(ed25519::ED25519_SIGNATURE_SIZE, 64);
+        let mut sig = [0u8; 64];
+        ed25519::sign(public, secret, digest, &mut sig)?;
+        Ok(sig)
+    }
+
+    fn ed25519_verify(public: &[u8; 32], digest: &[u8], signature: &[u8; 64])
+                      -> Result<bool> {
+        debug_assert_eq!(ed25519::ED25519_KEY_SIZE, 32);
+        debug_assert_eq!(ed25519::ED25519_SIGNATURE_SIZE, 64);
+        Ok(ed25519::verify(public, digest, signature)?)
+    }
+
+    fn dsa_generate_key(p_bits: usize)
+                        -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
+    {
+        let mut rng = Yarrow::default();
+        let q_bits = if p_bits <= 1024 { 160 } else { 256 };
+        let params = dsa::Params::generate(&mut rng, p_bits, q_bits)?;
+        let (p, q) = params.primes();
+        let g = params.g();
+        let (y, x) = dsa::generate_keypair(&params, &mut rng);
+        Ok((p.into(), q.into(), g.into(), y.as_bytes().into(),
+            x.as_bytes().into()))
+    }
+}
+
+impl KeyPair {
+    pub(crate) fn sign_backend(&self,
+                               secret: &mpi::SecretKeyMaterial,
+                               hash_algo: HashAlgorithm,
+                               digest: &[u8])
+                               -> Result<mpi::Signature>
     {
         use crate::PublicKeyAlgorithm::*;
 
         let mut rng = Yarrow::default();
 
-        self.secret().map(|secret| {
-            #[allow(deprecated)]
-            match (self.public().pk_algo(), self.public().mpis(), secret)
+        #[allow(deprecated)]
+        match (self.public().pk_algo(), self.public().mpis(), secret)
         {
             (RSASign,
              &PublicKey::RSA { ref e, ref n },
@@ -72,32 +164,6 @@ impl Signer for KeyPair {
                 })
             },
 
-            (EdDSA,
-             &PublicKey::EdDSA { ref curve, ref q },
-             &mpi::SecretKeyMaterial::EdDSA { ref scalar }) => match curve {
-                Curve::Ed25519 => {
-                    let public = q.decode_point(&Curve::Ed25519)?.0;
-
-                    let mut sig = vec![0; ed25519::ED25519_SIGNATURE_SIZE];
-
-                    // Nettle expects the private key to be exactly
-                    // ED25519_KEY_SIZE bytes long but OpenPGP allows leading
-                    // zeros to be stripped.
-                    // Padding has to be unconditional; otherwise we have a
-                    // secret-dependent branch.
-                    let sec = scalar.value_padded(ed25519::ED25519_KEY_SIZE);
-
-                    ed25519::sign(public, &sec[..], digest, &mut sig)?;
-
-                    Ok(mpi::Signature::EdDSA {
-                        r: MPI::new(&sig[..ed25519::ED25519_KEY_SIZE]),
-                        s: MPI::new(&sig[ed25519::ED25519_KEY_SIZE..]),
-                    })
-                },
-                _ => Err(
-                    Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
-
             (ECDSA,
              &PublicKey::ECDSA { ref curve, .. },
              &mpi::SecretKeyMaterial::ECDSA { ref scalar }) => {
@@ -129,24 +195,18 @@ impl Signer for KeyPair {
                 "unsupported combination of algorithm {:?}, key {:?}, \
                  and secret key {:?}",
                 pk_algo, self.public(), self.secret())).into()),
-        }})
+        }
     }
 }
 
-impl Decryptor for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn decrypt(&mut self, ciphertext: &mpi::Ciphertext,
+impl KeyPair {
+    pub(crate) fn decrypt_backend(&self, secret: &mpi::SecretKeyMaterial, ciphertext: &mpi::Ciphertext,
                plaintext_len: Option<usize>)
                -> Result<SessionKey>
     {
         use crate::PublicKeyAlgorithm::*;
 
-        self.secret().map(
-            |secret| Ok(match (self.public().mpis(), secret, ciphertext)
-        {
+        Ok(match (self.public().mpis(), secret, ciphertext) {
             (PublicKey::RSA{ ref e, ref n },
              mpi::SecretKeyMaterial::RSA{ ref p, ref q, ref d, .. },
              mpi::Ciphertext::RSA{ ref c }) => {
@@ -175,21 +235,22 @@ impl Decryptor for KeyPair {
             (PublicKey::ECDH{ .. },
              mpi::SecretKeyMaterial::ECDH { .. },
              mpi::Ciphertext::ECDH { .. }) =>
-                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext)?,
+                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext,
+                                             plaintext_len)?,
 
             (public, secret, ciphertext) =>
                 return Err(Error::InvalidOperation(format!(
                     "unsupported combination of key pair {:?}/{:?} \
                      and ciphertext {:?}",
                     public, secret, ciphertext)).into()),
-        }))
+        })
     }
 }
 
 
 impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     /// Encrypts the given data with this key.
-    pub fn encrypt(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
+    pub(crate) fn encrypt_backend(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
         use crate::PublicKeyAlgorithm::*;
 
         #[allow(deprecated)]
@@ -238,15 +299,11 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     }
 
     /// Verifies the given signature.
-    pub fn verify(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
+    pub(crate) fn verify_backend(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
                   digest: &[u8]) -> Result<()>
     {
         use crate::crypto::mpi::Signature;
 
-        fn bad(e: impl ToString) -> anyhow::Error {
-            Error::BadSignature(e.to_string()).into()
-        }
-
         let ok = match (self.mpis(), sig) {
             (PublicKey::RSA { e, n }, Signature::RSA { s }) => {
                 let key = rsa::PublicKey::new(n.value(), e.value())?;
@@ -267,37 +324,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
 
                 dsa::verify(&params, &key, digest, &signature)
             },
-            (PublicKey::EdDSA { curve, q }, Signature::EdDSA { r, s }) =>
-              match curve {
-                Curve::Ed25519 => {
-                    if q.value().get(0).map(|&b| b != 0x40).unwrap_or(true) {
-                        return Err(Error::MalformedPacket(
-                            "Invalid point encoding".into()).into());
-                    }
-
-                    // OpenPGP encodes R and S separately, but our
-                    // cryptographic library expects them to be
-                    // concatenated.
-                    let mut signature =
-                        Vec::with_capacity(ed25519::ED25519_SIGNATURE_SIZE);
-
-                    // We need to zero-pad them at the front, because
-                    // the MPI encoding drops leading zero bytes.
-                    let half = ed25519::ED25519_SIGNATURE_SIZE / 2;
-                    signature.extend_from_slice(
-                        &r.value_padded(half).map_err(bad)?);
-                    signature.extend_from_slice(
-                        &s.value_padded(half).map_err(bad)?);
-
-                    // Let's see if we got it right.
-                    assert_eq!(signature.len(),
-                               ed25519::ED25519_SIGNATURE_SIZE);
-
-                    ed25519::verify(&q.value()[1..], digest, &signature)?
-                },
-                _ => return
-                    Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
             (PublicKey::ECDSA { curve, q }, Signature::ECDSA { s, r }) =>
             {
                 let (x, y) = q.decode_point(curve)?;
@@ -333,37 +359,6 @@ use crate::types::PublicKeyAlgorithm;
 impl<R> Key4<SecretParts, R>
     where R: key::KeyRole,
 {
-    pub(crate) fn derive_cv25519_public_key(private_key: &Protected) -> Result<[u8; 32]>
-    {
-        let mut public_key = [0; ed25519::ED25519_KEY_SIZE];
-        curve25519::mul_g(&mut public_key, private_key)?;
-        Ok(public_key)
-    }
-
-    /// Creates a new OpenPGP secret key packet for an existing Ed25519 key.
-    ///
-    /// The ECDH key will use hash algorithm `hash` and symmetric
-    /// algorithm `sym`.  If one or both are `None` secure defaults
-    /// will be used.  The key will have it's creation date set to
-    /// `ctime` or the current time if `None` is given.
-    pub fn import_secret_ed25519<T>(private_key: &[u8], ctime: T)
-        -> Result<Self> where T: Into<Option<SystemTime>>
-    {
-        let mut public_key = [0; ed25519::ED25519_KEY_SIZE];
-        ed25519::public_key(&mut public_key, private_key).unwrap();
-
-        Self::with_secret(
-            ctime.into().unwrap_or_else(crate::now),
-            PublicKeyAlgorithm::EdDSA,
-            mpi::PublicKey::EdDSA {
-                curve: Curve::Ed25519,
-                q: MPI::new_compressed_point(&public_key),
-            },
-            mpi::SecretKeyMaterial::EdDSA {
-                scalar: private_key.into(),
-            }.into())
-    }
-
     /// Creates a new OpenPGP public key packet for an existing RSA key.
     ///
     /// The RSA key will use public exponent `e` and modulo `n`. The key will
@@ -392,7 +387,7 @@ impl<R> Key4<SecretParts, R>
             }.into())
     }
 
-    /// Generates a new RSA key with a public modulos of size `bits`.
+    /// Generates a new RSA key with a public modulus of size `bits`.
     pub fn generate_rsa(bits: usize) -> Result<Self> {
         let mut rng = Yarrow::default();
 
@@ -449,10 +444,8 @@ impl<R> Key4<SecretParts, R>
             }
 
             (Curve::Cv25519, false) => {
-                let mut public = [0; curve25519::CURVE25519_SIZE];
-                let mut private: Protected =
-                    curve25519::private_key(&mut rng).into();
-                curve25519::mul_g(&mut public, &private)?;
+                let (mut private, public) =
+                    super::Backend::x25519_generate_key()?;
 
                 // Reverse the scalar.  See
                 // https://lists.gnupg.org/pipermail/gnupg-devel/2018-February/033437.html.
