@@ -1,6 +1,7 @@
 use crate::{Error, Result};
 
-use crate::crypto::asymmetric::{Decryptor, KeyPair, Signer};
+use crate::crypto::asymmetric::KeyPair;
+use crate::crypto::backend::interface::Asymmetric;
 use crate::crypto::mpi;
 use crate::crypto::mpi::{ProtectedMPI, MPI};
 use crate::crypto::mem::Protected;
@@ -12,6 +13,7 @@ use std::convert::{TryFrom, TryInto};
 use std::time::SystemTime;
 
 use openssl::bn::{BigNum, BigNumRef, BigNumContext};
+use openssl::derive::Deriver;
 use openssl::ec::{EcGroup, EcKey, EcPoint, PointConversionForm};
 use openssl::ecdsa::EcdsaSig;
 use openssl::nid::Nid;
@@ -21,6 +23,98 @@ use openssl::rsa::{Padding, Rsa, RsaPrivateKeyBuilder};
 use openssl::sign::Signer as OpenSslSigner;
 use openssl::sign::Verifier;
 
+impl Asymmetric for super::Backend {
+    fn supports_algo(algo: PublicKeyAlgorithm) -> bool {
+        use PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        match algo {
+            RSAEncryptSign | RSAEncrypt | RSASign => true,
+            DSA => true,
+            ECDH | ECDSA | EdDSA => true,
+            ElGamalEncrypt | ElGamalEncryptSign |
+            Private(_) | Unknown(_)
+                => false,
+        }
+    }
+
+    fn supports_curve(curve: &Curve) -> bool {
+        if matches!(curve, Curve::Ed25519 | Curve::Cv25519) {
+            // 25519-based algorithms are special-cased and supported
+            true
+        } else {
+            // the rest of EC algorithms are supported via the same
+            // codepath
+            if let Ok(nid) = openssl::nid::Nid::try_from(curve) {
+                openssl::ec::EcGroup::from_curve_name(nid).is_ok()
+            } else {
+                false
+            }
+        }
+    }
+
+    fn x25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        let pair = openssl::pkey::PKey::generate_x25519()?;
+        Ok((pair.raw_private_key()?.into(),
+            pair.raw_public_key()?.as_slice().try_into()?))
+    }
+
+    fn x25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::X25519)?;
+        Ok(key.raw_public_key()?.as_slice().try_into()?)
+    }
+
+    fn x25519_shared_point(secret: &Protected, public: &[u8; 32])
+                           -> Result<Protected> {
+        let public = PKey::public_key_from_raw_bytes(
+            public, openssl::pkey::Id::X25519)?;
+        let secret = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::X25519)?;
+
+        let mut deriver = Deriver::new(&secret)?;
+        deriver.set_peer(&public)?;
+        Ok(deriver.derive_to_vec()?.into())
+    }
+
+    fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        let pair = openssl::pkey::PKey::generate_ed25519()?;
+        Ok((pair.raw_private_key()?.into(),
+            pair.raw_public_key()?.as_slice().try_into()?))
+    }
+
+    fn ed25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::ED25519)?;
+        Ok(key.raw_public_key()?.as_slice().try_into()?)
+    }
+
+    fn ed25519_sign(secret: &Protected, _public: &[u8; 32], digest: &[u8])
+                    -> Result<[u8; 64]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::ED25519)?;
+
+        let mut signer = OpenSslSigner::new_without_digest(&key)?;
+        Ok(signer.sign_oneshot_to_vec(digest)?.as_slice().try_into()?)
+    }
+
+    fn ed25519_verify(public: &[u8; 32], digest: &[u8], signature: &[u8; 64])
+                      -> Result<bool> {
+        let key = PKey::public_key_from_raw_bytes(
+            public, openssl::pkey::Id::ED25519)?;
+        let mut verifier = Verifier::new_without_digest(&key)?;
+        Ok(verifier.verify_oneshot(signature, digest)?)
+    }
+
+    fn dsa_generate_key(p_bits: usize)
+                        -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
+    {
+        use openssl::dsa::*;
+        let key = Dsa::<openssl::pkey::Private>::generate(p_bits.try_into()?)?;
+        Ok((key.p().into(), key.q().into(), key.g().into(),
+            key.pub_key().into(), key.priv_key().into()))
+    }
+}
+
 impl TryFrom<&ProtectedMPI> for BigNum {
     type Error = anyhow::Error;
     fn try_from(mpi: &ProtectedMPI) -> std::result::Result<BigNum, anyhow::Error> {
@@ -79,16 +173,16 @@ impl TryFrom<&Curve> for Nid {
     }
 }
 
-impl Signer for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8]) -> Result<mpi::Signature> {
+impl KeyPair {
+    pub(crate) fn sign_backend(&self,
+                               secret: &mpi::SecretKeyMaterial,
+                               hash_algo: HashAlgorithm,
+                               digest: &[u8])
+                               -> Result<mpi::Signature>
+    {
         use crate::PublicKeyAlgorithm::*;
         #[allow(deprecated)]
-        self.secret().map(
-            |secret| match (self.public().pk_algo(), self.public().mpis(), secret) {
+        match (self.public().pk_algo(), self.public().mpis(), secret) {
                 (
                     RSAEncryptSign,
                     mpi::PublicKey::RSA { e, n },
@@ -165,30 +259,6 @@ impl Signer for KeyPair {
                     })
                 }
 
-                (
-                    EdDSA,
-                    mpi::PublicKey::EdDSA { curve, .. },
-                    mpi::SecretKeyMaterial::EdDSA { scalar },
-                ) => match curve {
-                    Curve::Ed25519 => {
-                        let scalar = scalar.value_padded(32);
-
-                        let key =
-                            PKey::private_key_from_raw_bytes(&scalar, openssl::pkey::Id::ED25519)?;
-
-                        let mut signer = OpenSslSigner::new_without_digest(&key)?;
-                        let signature = signer.sign_oneshot_to_vec(digest)?;
-
-                        // https://tools.ietf.org/html/rfc8032#section-5.1.6
-                        let (r, s) = signature.split_at(signature.len() / 2);
-                        Ok(mpi::Signature::EdDSA {
-                            r: r.to_vec().into(),
-                            s: s.to_vec().into(),
-                        })
-                    }
-                    _ => Err(crate::Error::UnsupportedEllipticCurve(curve.clone()).into()),
-                },
-
                 (pk_algo, _, _) => Err(crate::Error::InvalidOperation(format!(
                     "unsupported combination of algorithm {:?}, key {:?}, \
                         and secret key {:?} by OpenSSL backend",
@@ -197,25 +267,20 @@ impl Signer for KeyPair {
                     self.secret()
                 ))
                 .into()),
-            },
-        )
+        }
     }
 }
 
-impl Decryptor for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn decrypt(
-        &mut self,
+impl KeyPair {
+    pub(crate) fn decrypt_backend(
+        &self,
+        secret: &mpi::SecretKeyMaterial,
         ciphertext: &mpi::Ciphertext,
-        _plaintext_len: Option<usize>,
+        plaintext_len: Option<usize>,
     ) -> Result<SessionKey> {
         use crate::crypto::mpi::PublicKey;
 
-        self.secret().map(|secret| {
-            Ok(match (self.public().mpis(), secret, ciphertext) {
+        Ok(match (self.public().mpis(), secret, ciphertext) {
                 (
                     PublicKey::RSA { ref e, ref n },
                     mpi::SecretKeyMaterial::RSA {
@@ -240,7 +305,9 @@ impl Decryptor for KeyPair {
                     PublicKey::ECDH { .. },
                     mpi::SecretKeyMaterial::ECDH { .. },
                     mpi::Ciphertext::ECDH { .. },
-                ) => crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext)?,
+                ) => crate::crypto::ecdh::decrypt(self.public(), secret,
+                                                  ciphertext,
+                                                  plaintext_len)?,
 
                 (public, secret, ciphertext) => {
                     return Err(crate::Error::InvalidOperation(format!(
@@ -250,14 +317,13 @@ impl Decryptor for KeyPair {
                     ))
                     .into())
                 }
-            })
         })
     }
 }
 
 impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     /// Encrypts the given data with this key.
-    pub fn encrypt(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
+    pub(crate) fn encrypt_backend(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
         use PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match self.pk_algo() {
@@ -304,7 +370,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     }
 
     /// Verifies the given signature.
-    pub fn verify(
+    pub(crate) fn verify_backend(
         &self,
         sig: &mpi::Signature,
         hash_algo: HashAlgorithm,
@@ -342,29 +408,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                 ctx.verify_init()?;
                 ctx.verify(&digest, &signature.to_der()?)?
             }
-            (mpi::PublicKey::EdDSA { curve, q }, mpi::Signature::EdDSA { r, s }) => match curve {
-                Curve::Ed25519 => {
-                    let public = q.decode_point(&Curve::Ed25519)?.0;
-
-                    let key = PKey::public_key_from_raw_bytes(public, openssl::pkey::Id::ED25519)?;
-
-                    const SIGNATURE_LENGTH: usize = 64;
-
-                    // ed25519 expects full-sized signatures but OpenPGP allows
-                    // for stripped leading zeroes, pad each part with zeroes.
-                    let mut sig_bytes = [0u8; SIGNATURE_LENGTH];
-
-                    // We need to zero-pad them at the front, because
-                    // the MPI encoding drops leading zero bytes.
-                    let half = SIGNATURE_LENGTH / 2;
-                    sig_bytes[..half].copy_from_slice(&r.value_padded(half)?);
-                    sig_bytes[half..].copy_from_slice(&s.value_padded(half)?);
-
-                    let mut verifier = Verifier::new_without_digest(&key)?;
-                    verifier.verify_oneshot(&sig_bytes, digest)?
-                }
-                _ => return Err(crate::Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
             (mpi::PublicKey::ECDSA { curve, q }, mpi::Signature::ECDSA { s, r }) => {
                 let nid = curve.try_into()?;
                 let group = EcGroup::from_curve_name(nid)?;
@@ -399,40 +442,6 @@ impl<R> Key4<SecretParts, R>
 where
     R: key::KeyRole,
 {
-    pub(crate) fn derive_cv25519_public_key(private_key: &Protected) -> Result<[u8; 32]>
-    {
-        let key = PKey::private_key_from_raw_bytes(private_key, openssl::pkey::Id::X25519)?;
-        let public = key.raw_public_key()?;
-        Ok(<[u8; 32]>::try_from(&public[..])?)
-    }
-
-    /// Creates a new OpenPGP secret key packet for an existing Ed25519 key.
-    ///
-    /// The ECDH key will use hash algorithm `hash` and symmetric
-    /// algorithm `sym`.  If one or both are `None` secure defaults
-    /// will be used.  The key will have it's creation date set to
-    /// `ctime` or the current time if `None` is given.
-    pub fn import_secret_ed25519<T>(private_key: &[u8], ctime: T) -> Result<Self>
-    where
-        T: Into<Option<SystemTime>>,
-    {
-        let key = PKey::private_key_from_raw_bytes(private_key, openssl::pkey::Id::ED25519)?;
-        let public_key = key.raw_public_key()?;
-
-        Self::with_secret(
-            ctime.into().unwrap_or_else(crate::now),
-            PublicKeyAlgorithm::EdDSA,
-            mpi::PublicKey::EdDSA {
-                curve: Curve::Ed25519,
-                q: public_key.into(),
-            },
-            mpi::SecretKeyMaterial::EdDSA {
-                scalar: private_key.into(),
-            }
-            .into(),
-        )
-    }
-
     /// Creates a new OpenPGP public key packet for an existing RSA key.
     ///
     /// The RSA key will use public exponent `e` and modulo `n`. The key will
@@ -483,7 +492,7 @@ where
         )
     }
 
-    /// Generates a new RSA key with a public modulos of size `bits`.
+    /// Generates a new RSA key with a public modulus of size `bits`.
     #[allow(clippy::many_single_char_names)]
     pub fn generate_rsa(bits: usize) -> Result<Self> {
         let key = Rsa::generate(bits.try_into()?)?;
