@@ -7,13 +7,13 @@
 use std::convert::TryFrom;
 use std::time::SystemTime;
 
-use x25519_dalek_ng as x25519_dalek;
 use num_bigint_dig::{traits::ModInverse, BigUint};
 use rsa::traits::{PrivateKeyParts, PublicKeyParts};
 use rsa::{Pkcs1v15Encrypt, RsaPublicKey, RsaPrivateKey, Pkcs1v15Sign};
 
 use crate::{Error, Result};
-use crate::crypto::asymmetric::{KeyPair, Decryptor, Signer};
+use crate::crypto::asymmetric::KeyPair;
+use crate::crypto::backend::interface::Asymmetric;
 use crate::crypto::mem::Protected;
 use crate::crypto::mpi::{self, MPI, ProtectedMPI};
 use crate::crypto::SessionKey;
@@ -26,10 +26,171 @@ use super::GenericArrayExt;
 
 const CURVE25519_SIZE: usize = 32;
 
+impl TryFrom<&Protected> for Box<ed25519_dalek::SigningKey> {
+    type Error = anyhow::Error;
+
+    fn try_from(value: &Protected) -> Result<Self> {
+        if value.len() != ed25519_dalek::SECRET_KEY_LENGTH {
+            return Err(crate::Error::InvalidArgument(
+                "Bad Ed25519 secret length".into()).into());
+        }
+        Ok(Box::new(ed25519_dalek::SigningKey::from_bytes(
+            value.as_ref().try_into().map_err(
+                |e: std::array::TryFromSliceError| {
+                    Error::InvalidKey(e.to_string())
+                })?)))
+    }
+}
+
+impl Asymmetric for super::Backend {
+    fn supports_algo(algo: PublicKeyAlgorithm) -> bool {
+        use PublicKeyAlgorithm::*;
+        #[allow(deprecated)]
+        match algo {
+            RSAEncryptSign | RSAEncrypt | RSASign | ECDH | EdDSA | ECDSA
+                => true,
+            DSA
+                => true,
+            ElGamalEncrypt | ElGamalEncryptSign | Private(_) | Unknown(_)
+                => false,
+        }
+    }
+
+    fn supports_curve(curve: &Curve) -> bool {
+        use self::Curve::*;
+        match curve {
+            NistP256
+                => true,
+            NistP384 | NistP521
+                => false,
+            Ed25519 | Cv25519
+                => true,
+            BrainpoolP256 | BrainpoolP512 | Unknown(_)
+                => false,
+        }
+    }
+
+    fn x25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        use x25519_dalek::{StaticSecret, PublicKey};
+
+        // x25519_dalek v1.1 doesn't reexport OsRng.  It
+        // depends on rand 0.8.
+        use rand::rngs::OsRng;
+
+        let secret = StaticSecret::random_from_rng(&mut OsRng);
+        let public = PublicKey::from(&secret);
+        let mut secret_bytes = secret.to_bytes();
+        let secret: Protected = secret_bytes.as_ref().into();
+        unsafe {
+            memsec::memzero(secret_bytes.as_mut_ptr(), secret_bytes.len());
+        }
+
+        Ok((secret, public.to_bytes()))
+    }
+
+    fn x25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        use x25519_dalek::{PublicKey, StaticSecret};
+
+        let secret = StaticSecret::from(<[u8; 32]>::try_from(&secret[..])?);
+        Ok(*PublicKey::from(&secret).as_bytes())
+    }
+
+    fn x25519_shared_point(secret: &Protected, public: &[u8; 32])
+                           -> Result<Protected> {
+        use x25519_dalek::{StaticSecret, PublicKey};
+
+        let secret = StaticSecret::from(<[u8; 32]>::try_from(&secret[..])?);
+        let public = PublicKey::from(public.clone());
+        Ok((&secret.diffie_hellman(&public).as_bytes()[..]).into())
+    }
+
+    fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
+        use rand::rngs::OsRng as OsRng;
+        let pair = ed25519_dalek::SigningKey::generate(&mut OsRng);
+        Ok((pair.to_bytes().into(), pair.verifying_key().to_bytes()))
+    }
+
+    fn ed25519_derive_public(secret: &Protected) -> Result<[u8; 32]> {
+        use ed25519_dalek::SigningKey;
+        let secret: Box<SigningKey> = secret.try_into()?;
+        let public = secret.verifying_key();
+        Ok(public.to_bytes())
+    }
+
+    fn ed25519_sign(secret: &Protected, _public: &[u8; 32], digest: &[u8])
+                    -> Result<[u8; 64]> {
+        use ed25519_dalek::{SigningKey, Signer};
+        let pair: Box<SigningKey> = secret.try_into()?;
+        Ok(pair.sign(digest).to_bytes().try_into()?)
+    }
+
+    fn ed25519_verify(public: &[u8; 32], digest: &[u8], signature: &[u8; 64])
+                      -> Result<bool> {
+        use ed25519_dalek::{VerifyingKey, Verifier, Signature};
+
+        let public = VerifyingKey::from_bytes(public).map_err(|e| {
+            Error::InvalidKey(e.to_string())
+        })?;
+        let signature = signature.as_ref().try_into().map_err(|e: std::array::TryFromSliceError| {
+            Error::InvalidArgument(e.to_string())
+        })?;
+
+        let signature = Signature::from_bytes(signature);
+        Ok(public.verify(digest, &signature).is_ok())
+    }
+
+    fn dsa_generate_key(p_bits: usize)
+                        -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
+    {
+        #[allow(deprecated)]
+        let size = match p_bits {
+            1024 => dsa::KeySize::DSA_1024_160,
+            2048 => dsa::KeySize::DSA_2048_256,
+            3072 => dsa::KeySize::DSA_3072_256,
+            n => return Err(Error::InvalidArgument(
+                format!("Key size {} is not supported", n)).into()),
+        };
+
+        let mut rng = rand_core::OsRng;
+        let components = dsa::Components::generate(&mut rng, size);
+        let p = components.p().into();
+        let q = components.q().into();
+        let g = components.g().into();
+        let secret = dsa::SigningKey::generate(&mut rng, components);
+        let public = secret.verifying_key();
+
+        Ok((p, q, g, public.y().into(), secret.x().into()))
+    }
+}
+
+impl From<&BigUint> for ProtectedMPI {
+    fn from(v: &BigUint) -> Self {
+        v.to_bytes_be().into()
+    }
+}
+
+impl From<&ProtectedMPI> for BigUint {
+    fn from(v: &ProtectedMPI) -> Self {
+        BigUint::from_bytes_be(v.value()).into()
+    }
+}
+
+impl From<&BigUint> for MPI {
+    fn from(v: &BigUint) -> Self {
+        v.to_bytes_be().into()
+    }
+}
+
+impl From<&MPI> for BigUint {
+    fn from(v: &MPI) -> Self {
+        BigUint::from_bytes_be(v.value()).into()
+    }
+}
+
 fn pkcs1_padding(hash_algo: HashAlgorithm) -> Result<Pkcs1v15Sign> {
     let hash = match hash_algo {
         HashAlgorithm::MD5 => Pkcs1v15Sign::new::<md5::Md5>(),
-        HashAlgorithm::SHA1 => Pkcs1v15Sign::new::<sha1::Sha1>(),
+        HashAlgorithm::SHA1 => Pkcs1v15Sign::new::<sha1collisiondetection::Sha1CD>(),
         HashAlgorithm::SHA224 => Pkcs1v15Sign::new::<sha2::Sha224>(),
         HashAlgorithm::SHA256 => Pkcs1v15Sign::new::<sha2::Sha256>(),
         HashAlgorithm::SHA384 => Pkcs1v15Sign::new::<sha2::Sha384>(),
@@ -59,17 +220,16 @@ fn rsa_private_key(e: &MPI, n: &MPI, p: &ProtectedMPI, q: &ProtectedMPI, d: &Pro
     Ok(RsaPrivateKey::from_components(n, e, d, vec![p, q])?)
 }
 
-impl Signer for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn sign(&mut self, hash_algo: HashAlgorithm, digest: &[u8])
+impl KeyPair {
+    pub(crate) fn sign_backend(&self,
+                               secret: &mpi::SecretKeyMaterial,
+                               hash_algo: HashAlgorithm,
+                               digest: &[u8])
             -> Result<mpi::Signature>
     {
         use crate::PublicKeyAlgorithm::*;
         #[allow(deprecated)]
-        self.secret().map(|secret| match (self.public().pk_algo(), self.public().mpis(), secret) {
+        match (self.public().pk_algo(), self.public().mpis(), secret) {
             (RSAEncryptSign,
              mpi::PublicKey::RSA { e, n },
              mpi::SecretKeyMaterial::RSA { p, q, d, .. }) |
@@ -85,9 +245,20 @@ impl Signer for KeyPair {
             },
 
             (PublicKeyAlgorithm::DSA,
-             mpi:: PublicKey::DSA { .. },
-             mpi::SecretKeyMaterial::DSA { .. }) => {
-                Err(Error::UnsupportedPublicKeyAlgorithm(PublicKeyAlgorithm::DSA).into())
+             mpi::PublicKey::DSA { p, q, g, y },
+             mpi::SecretKeyMaterial::DSA { x }) => {
+                use dsa::signature::hazmat::PrehashSigner;
+                let c = dsa::Components::from_components(
+                    p.into(), q.into(), g.into())?;
+                let public =
+                    dsa::VerifyingKey::from_components(c, y.into())?;
+                let secret =
+                    dsa::SigningKey::from_components(public, x.into())?;
+                let sig = secret.sign_prehash(digest)?;
+                Ok(mpi::Signature::DSA {
+                    r: sig.r().into(),
+                    s: sig.s().into(),
+                })
             },
 
             (PublicKeyAlgorithm::ECDSA,
@@ -130,44 +301,6 @@ impl Signer for KeyPair {
                 _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
             },
 
-            (EdDSA,
-             mpi::PublicKey::EdDSA { curve, q },
-             mpi::SecretKeyMaterial::EdDSA { scalar }) => match curve
-            {
-                Curve::Ed25519 => {
-                    use ed25519_dalek::{Keypair, Signer};
-                    use ed25519_dalek::{PUBLIC_KEY_LENGTH, SECRET_KEY_LENGTH};
-
-                    let (public, ..) = q.decode_point(&Curve::Ed25519)?;
-                    assert_eq!(public.len(), PUBLIC_KEY_LENGTH);
-
-                    // It's expected for the private key to be exactly
-                    // SECRET_KEY_LENGTH bytes long but OpenPGP allows leading
-                    // zeros to be stripped.
-                    // Padding has to be unconditional; otherwise we have a
-                    // secret-dependent branch.
-                    let mut keypair = Protected::from(
-                        vec![0u8; SECRET_KEY_LENGTH + PUBLIC_KEY_LENGTH]
-                    );
-                    keypair.as_mut()[..SECRET_KEY_LENGTH]
-                        .copy_from_slice(
-                            &scalar.value_padded(SECRET_KEY_LENGTH));
-                    keypair.as_mut()[SECRET_KEY_LENGTH..]
-                        .copy_from_slice(public);
-                    let pair = Keypair::from_bytes(&keypair)?;
-
-                    let sig = pair.sign(digest).to_bytes();
-
-                    // https://tools.ietf.org/html/rfc8032#section-5.1.6
-                    let (r, s) = sig.split_at(sig.len() / 2);
-                    Ok(mpi::Signature::EdDSA {
-                        r: mpi::MPI::new(r),
-                        s: mpi::MPI::new(s),
-                    })
-                },
-                _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
-            },
-
             (pk_algo, _, _) => {
                 Err(Error::InvalidOperation(format!(
                     "unsupported combination of algorithm {:?}, key {:?}, \
@@ -177,21 +310,17 @@ impl Signer for KeyPair {
                     self.secret()
                 )).into())
             }
-        })
+        }
     }
 }
 
-impl Decryptor for KeyPair {
-    fn public(&self) -> &Key<key::PublicParts, key::UnspecifiedRole> {
-        KeyPair::public(self)
-    }
-
-    fn decrypt(&mut self, ciphertext: &mpi::Ciphertext,
-               _plaintext_len: Option<usize>)
+impl KeyPair {
+    pub(crate) fn decrypt_backend(&self, secret: &mpi::SecretKeyMaterial, ciphertext: &mpi::Ciphertext,
+               plaintext_len: Option<usize>)
                -> Result<SessionKey>
     {
         use crate::PublicKeyAlgorithm::*;
-        self.secret().map(|secret| match (self.public().mpis(), secret, ciphertext) {
+        match (self.public().mpis(), secret, ciphertext) {
             (mpi::PublicKey::RSA { e, n },
              mpi::SecretKeyMaterial::RSA { p, q, d, .. },
              mpi::Ciphertext::RSA { c }) => {
@@ -208,21 +337,22 @@ impl Decryptor for KeyPair {
             (mpi::PublicKey::ECDH { .. },
              mpi::SecretKeyMaterial::ECDH { .. },
              mpi::Ciphertext::ECDH { .. }) =>
-                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext),
+                crate::crypto::ecdh::decrypt(self.public(), secret, ciphertext,
+                                             plaintext_len),
 
             (public, secret, ciphertext) =>
                 Err(Error::InvalidOperation(format!(
                     "unsupported combination of key pair {:?}/{:?} \
                      and ciphertext {:?}",
                     public, secret, ciphertext)).into()),
-        })
+        }
     }
 }
 
 
 impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     /// Encrypts the given data with this key.
-    pub fn encrypt(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
+    pub(crate) fn encrypt_backend(&self, data: &SessionKey) -> Result<mpi::Ciphertext> {
         use PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match self.pk_algo() {
@@ -260,7 +390,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
     }
 
     /// Verifies the given signature.
-    pub fn verify(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
+    pub(crate) fn verify_backend(&self, sig: &mpi::Signature, hash_algo: HashAlgorithm,
                   digest: &[u8]) -> Result<()>
     {
         fn bad(e: impl ToString) -> anyhow::Error {
@@ -288,9 +418,15 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                 key.verify(padding, digest, &s.value_padded(key.size())?)?;
                 Ok(())
             }
-            (mpi::PublicKey::DSA { .. },
-             mpi::Signature::DSA { .. }) => {
-                Err(Error::UnsupportedPublicKeyAlgorithm(PublicKeyAlgorithm::DSA).into())
+            (mpi::PublicKey::DSA { p, q, g, y },
+             mpi::Signature::DSA { r, s }) => {
+                use dsa::signature::hazmat::PrehashVerifier;
+                let c = dsa::Components::from_components(
+                    p.into(), q.into(), g.into())?;
+                let public = dsa::VerifyingKey::from_components(c, y.into())?;
+                let sig = dsa::Signature::from_components(r.into(), s.into())?;
+                public.verify_prehash(digest, &sig)?;
+                Ok(())
             },
             (mpi::PublicKey::ECDSA { curve, q },
              mpi::Signature::ECDSA { r, s }) => match curve
@@ -331,40 +467,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                 },
                 _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
             },
-            (mpi::PublicKey::EdDSA { curve, q },
-             mpi::Signature::EdDSA { r, s }) => match curve {
-                Curve::Ed25519 => {
-                    use ed25519_dalek::{PublicKey, Signature, SIGNATURE_LENGTH};
-                    use ed25519_dalek::{Verifier};
-
-                    let (public, ..) = q.decode_point(&Curve::Ed25519)?;
-                    assert_eq!(public.len(), 32);
-
-                    let key = PublicKey::from_bytes(public).map_err(|e| {
-                        Error::InvalidKey(e.to_string())
-                    })?;
-
-                    // OpenPGP encodes R and S separately, but our
-                    // cryptographic library expects them to be
-                    // concatenated.
-                    let mut sig_bytes = [0u8; SIGNATURE_LENGTH];
-
-                    // We need to zero-pad them at the front, because
-                    // the MPI encoding drops leading zero bytes.
-                    let half = SIGNATURE_LENGTH / 2;
-                    sig_bytes[..half].copy_from_slice(
-                        &r.value_padded(half).map_err(bad)?);
-                    sig_bytes[half..].copy_from_slice(
-                        &s.value_padded(half).map_err(bad)?);
-
-                    let signature = Signature::from(sig_bytes);
-
-                    key.verify(digest, &signature)
-                        .map_err(|e| Error::BadSignature(e.to_string()))?;
-                    Ok(())
-                },
-                _ => Err(Error::UnsupportedEllipticCurve(curve.clone()).into()),
-        },
             _ => Err(Error::MalformedPacket(format!(
                 "unsupported combination of key {} and signature {:?}.",
                 self.pk_algo(), sig)).into()),
@@ -375,48 +477,6 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
 impl<R> Key4<SecretParts, R>
     where R: key::KeyRole,
 {
-    pub(crate) fn derive_cv25519_public_key(private_key: &Protected) -> Result<[u8; 32]>
-    {
-        use x25519_dalek::{PublicKey, StaticSecret};
-
-        let secret = StaticSecret::from(<[u8; 32]>::try_from(&private_key[..])?);
-        Ok(*PublicKey::from(&secret).as_bytes())
-    }
-
-    /// Creates a new OpenPGP secret key packet for an existing Ed25519 key.
-    ///
-    /// The ECDH key will use hash algorithm `hash` and symmetric
-    /// algorithm `sym`.  If one or both are `None` secure defaults
-    /// will be used.  The key will have it's creation date set to
-    /// `ctime` or the current time if `None` is given.
-    pub fn import_secret_ed25519<T>(private_key: &[u8], ctime: T)
-        -> Result<Self> where T: Into<Option<SystemTime>>
-    {
-        use ed25519_dalek::{PublicKey, SecretKey};
-
-        let private = SecretKey::from_bytes(private_key).map_err(|e| {
-            Error::InvalidKey(e.to_string())
-        })?;
-
-        // Mark MPI as compressed point with 0x40 prefix. See
-        // https://tools.ietf.org/html/draft-ietf-openpgp-rfc4880bis-07#section-13.2.
-        let mut public = [0u8; 1 + CURVE25519_SIZE];
-        public[0] = 0x40;
-        public[1..].copy_from_slice(Into::<PublicKey>::into(&private).as_bytes());
-
-        Self::with_secret(
-            ctime.into().unwrap_or_else(crate::now),
-            PublicKeyAlgorithm::EdDSA,
-            mpi::PublicKey::EdDSA {
-                curve: Curve::Ed25519,
-                q: mpi::MPI::new(&public)
-            },
-            mpi::SecretKeyMaterial::EdDSA {
-                scalar: private_key.into(),
-            }.into()
-        )
-    }
-
     /// Creates a new OpenPGP public key packet for an existing RSA key.
     ///
     /// The RSA key will use public exponent `e` and modulo `n`. The key will
@@ -460,7 +520,7 @@ impl<R> Key4<SecretParts, R>
         )
     }
 
-    /// Generates a new RSA key with a public modulos of size `bits`.
+    /// Generates a new RSA key with a public modulus of size `bits`.
     pub fn generate_rsa(bits: usize) -> Result<Self> {
         let key = RsaPrivateKey::new(&mut rsa::rand_core::OsRng, bits)?;
         let (p, q) = match key.primes() {
@@ -503,22 +563,19 @@ impl<R> Key4<SecretParts, R>
 
         let (algo, public, private) = match (&curve, for_signing) {
             (Curve::Ed25519, true) => {
-                use ed25519_dalek::Keypair;
+                use ed25519_dalek::SigningKey;
 
-                // ed25519_dalek v1.0.1 doesn't reexport OsRng.  It
-                // depends on 0.7.
-                use rand07::rngs::OsRng as OsRng;
+                use rand::rngs::OsRng as OsRng;
 
-                let Keypair { public, secret }
-                    = Keypair::generate(&mut OsRng);
+                let key = SigningKey::generate(&mut OsRng);
 
-                let secret: Protected = secret.as_bytes().as_ref().into();
+                let secret: Protected = key.to_bytes().as_ref().into();
 
                 // Mark MPI as compressed point with 0x40 prefix. See
                 // https://tools.ietf.org/html/draft-ietf-openpgp-rfc4880bis-07#section-13.2.
                 let mut compressed_public = [0u8; 1 + CURVE25519_SIZE];
                 compressed_public[0] = 0x40;
-                compressed_public[1..].copy_from_slice(public.as_bytes());
+                compressed_public[1..].copy_from_slice(key.verifying_key().as_bytes());
 
                 (
                     PublicKeyAlgorithm::EdDSA,
@@ -528,26 +585,30 @@ impl<R> Key4<SecretParts, R>
             }
 
             (Curve::Cv25519, false) => {
-                use x25519_dalek::{StaticSecret, PublicKey};
+                let (mut private, public) =
+                    super::Backend::x25519_generate_key()?;
 
-                // x25519_dalek v1.1 doesn't reexport OsRng.  It
-                // depends on rand 0.8.
-                use rand::rngs::OsRng;
+                // x25519-dalek since v 2.0.0-rc.3 does not return clamped
+                // integers from Static Secrets but clamps them on usage.
+                // See: https://github.com/dalek-cryptography/x25519-dalek/blob/main/CHANGELOG.md#200-rc3
+                //
+                // Clamp the scalar.  X25519 does the clamping implicitly, but
+                // OpenPGP's ECDH over Curve25519 requires the secret to be
+                // clamped.
+                private[0] &= 0b1111_1000;
+                private[31] &= !0b1000_0000;
+                private[31] |= 0b0100_0000;
 
-                let private_key = StaticSecret::new(&mut OsRng);
-                let public_key = PublicKey::from(&private_key);
-
-                let mut private_key = Vec::from(private_key.to_bytes());
-                private_key.reverse();
+                private.reverse();
 
                 let public_mpis = mpi::PublicKey::ECDH {
                     curve: Curve::Cv25519,
-                    q: MPI::new_compressed_point(&*public_key.as_bytes()),
+                    q: MPI::new_compressed_point(&public),
                     hash,
                     sym,
                 };
                 let private_mpis = mpi::SecretKeyMaterial::ECDH {
-                    scalar: private_key.into(),
+                    scalar: private.into(),
                 };
 
                 (PublicKeyAlgorithm::ECDH, public_mpis, private_mpis)
