@@ -53,7 +53,7 @@ use crate::Error;
 use crate::Result;
 
 mod bitfield;
-pub(crate) use bitfield::Bitfield;
+pub use bitfield::Bitfield;
 mod compression_level;
 pub use compression_level::CompressionLevel;
 mod features;
@@ -207,7 +207,8 @@ impl PublicKeyAlgorithm {
     /// assert!(!PublicKeyAlgorithm::Private(101).is_supported());
     /// ```
     pub fn is_supported(&self) -> bool {
-        self.is_supported_by_backend()
+        use crate::crypto::backend::{Backend, interface::Asymmetric};
+        Backend::supports_algo(*self)
     }
 
     /// Returns an iterator over all valid variants.
@@ -380,6 +381,8 @@ impl Curve {
     /// For the Kobliz curves this is the size of the underlying
     /// finite field.  For X25519 it is 256.
     ///
+    /// This value is also equal to the length of a coordinate in bits.
+    ///
     /// Note: This information is useless and should not be used to
     /// gauge the security of a particular curve. This function exists
     /// only because some legacy PGP application like HKP need it.
@@ -457,6 +460,40 @@ impl Curve {
 impl fmt::Display for Curve {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         use self::Curve::*;
+
+        struct DotEncoded<'o>(&'o [u8]);
+        impl fmt::Display for DotEncoded<'_> {
+            fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+                let mut oid = self.0;
+                if oid.is_empty() {
+                    write!(f, "[invalid]")?;
+                    return Ok(());
+                }
+
+                // The first octet encodes two values.
+                let first = oid[0] / 40;
+                let second = oid[0] % 40;
+                oid = &oid[1..];
+                write!(f, "{}.{}", first, second)?;
+
+                let mut acc: usize = 0;
+                for b in oid {
+                    if b & 0x80 > 0 {
+                        acc *= 0x80;
+                        acc += (b & 0x7f) as usize;
+                    } else {
+                        acc *= 0x80;
+                        acc += (b & 0x7f) as usize;
+                        write!(f, ".{}", acc)?;
+                        acc = 0;
+                    }
+                }
+
+                Ok(())
+            }
+        }
+
+
         if f.alternate() {
             match *self {
                 NistP256 => f.write_str("NIST curve P-256"),
@@ -471,7 +508,7 @@ impl fmt::Display for Curve {
                 Cv25519
                     => f.write_str("Elliptic curve Diffie-Hellman using D.J. Bernstein's Curve25519"),
                 Unknown(ref oid)
-                    => write!(f, "Unknown curve (OID: {:?})", oid),
+                    => write!(f, "Unknown curve (OID: {})", DotEncoded(oid)),
             }
         } else {
             match *self {
@@ -487,7 +524,7 @@ impl fmt::Display for Curve {
                 Cv25519
                     => f.write_str("Curve25519"),
                 Unknown(ref oid)
-                    => write!(f, "Unknown curve {:?}", oid),
+                    => write!(f, "Unknown curve {}", DotEncoded(oid)),
             }
         }
     }
@@ -577,20 +614,11 @@ impl Curve {
     ///
     /// Returns `Error::UnsupportedEllipticCurve` if the curve is not
     /// supported.
+    #[deprecated(note = "Use bits()", since = "1.17.0")]
     pub fn len(&self) -> Result<usize> {
-        match self {
-            Curve::NistP256 => Ok(256),
-            Curve::NistP384 => Ok(384),
-            Curve::NistP521 => Ok(521),
-            Curve::BrainpoolP256 => Ok(256),
-            Curve::Unknown(_) if self.is_brainpoolp384() => Ok(384),
-            Curve::BrainpoolP512 => Ok(512),
-            Curve::Ed25519 => Ok(256),
-            Curve::Cv25519 => Ok(256),
-            Curve::Unknown(_) =>
-                Err(Error::UnsupportedEllipticCurve(self.clone())
-                    .into()),
-        }
+        self.bits()
+            .ok_or_else(|| Error::UnsupportedEllipticCurve(self.clone()).into())
+
     }
 
     /// Returns whether this algorithm is supported.
@@ -605,7 +633,8 @@ impl Curve {
     /// assert!(!Curve::Unknown(Box::new([0x2B, 0x11])).is_supported());
     /// ```
     pub fn is_supported(&self) -> bool {
-        self.is_supported_by_backend()
+        use crate::crypto::backend::{Backend, interface::Asymmetric};
+        Backend::supports_curve(self)
     }
 }
 
@@ -965,6 +994,8 @@ pub enum AEADAlgorithm {
     EAX,
     /// OCB mode.
     OCB,
+    /// Galois/Counter mode.
+    GCM,
     /// Private algorithm identifier.
     Private(u8),
     /// Unknown algorithm identifier.
@@ -972,11 +1003,18 @@ pub enum AEADAlgorithm {
 }
 assert_send_and_sync!(AEADAlgorithm);
 
-const AEAD_ALGORITHM_VARIANTS: [AEADAlgorithm; 2] = [
+const AEAD_ALGORITHM_VARIANTS: [AEADAlgorithm; 3] = [
     AEADAlgorithm::EAX,
     AEADAlgorithm::OCB,
+    AEADAlgorithm::GCM,
 ];
 
+impl Default for AEADAlgorithm {
+    fn default() -> Self {
+        Self::const_default()
+    }
+}
+
 impl AEADAlgorithm {
     /// Returns whether this algorithm is supported.
     ///
@@ -1007,6 +1045,7 @@ impl From<u8> for AEADAlgorithm {
         match u {
             1 => AEADAlgorithm::EAX,
             2 => AEADAlgorithm::OCB,
+            3 => AEADAlgorithm::GCM,
             100..=110 => AEADAlgorithm::Private(u),
             u => AEADAlgorithm::Unknown(u),
         }
@@ -1018,6 +1057,7 @@ impl From<AEADAlgorithm> for u8 {
         match s {
             AEADAlgorithm::EAX => 1,
             AEADAlgorithm::OCB => 2,
+            AEADAlgorithm::GCM => 3,
             AEADAlgorithm::Private(u) => u,
             AEADAlgorithm::Unknown(u) => u,
         }
@@ -1050,6 +1090,8 @@ impl fmt::Display for AEADAlgorithm {
                     f.write_str("EAX mode"),
                 AEADAlgorithm::OCB =>
                     f.write_str("OCB mode"),
+                AEADAlgorithm::GCM =>
+                    f.write_str("GCM mode"),
                 AEADAlgorithm::Private(u) =>
                     f.write_fmt(format_args!("Private/Experimental AEAD algorithm {}", u)),
                 AEADAlgorithm::Unknown(u) =>
@@ -1061,6 +1103,8 @@ impl fmt::Display for AEADAlgorithm {
                     f.write_str("EAX"),
                 AEADAlgorithm::OCB =>
                     f.write_str("OCB"),
+                AEADAlgorithm::GCM =>
+                    f.write_str("GCM"),
                 AEADAlgorithm::Private(u) =>
                     f.write_fmt(format_args!("Private AEAD algo {}", u)),
                 AEADAlgorithm::Unknown(u) =>
@@ -1222,7 +1266,7 @@ impl fmt::Display for CompressionAlgorithm {
             CompressionAlgorithm::Private(u) =>
                 f.write_fmt(format_args!("Private/Experimental compression algorithm {}", u)),
             CompressionAlgorithm::Unknown(u) =>
-                f.write_fmt(format_args!("Unknown comppression algorithm {}", u)),
+                f.write_fmt(format_args!("Unknown compression algorithm {}", u)),
         }
     }
 }
